#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <linux/i2c.h>

char *dev="/dev/i2c-0";
int  wr=0, x=0, adr=0xa0;

/* ---------------------------------------------------------------------- */

void usage(char *name)
{
    fprintf(stderr,"This is a i2c EEPROM tool\n");
    fprintf(stderr,"  read  data: %s [ options ] > file\n",name);
    fprintf(stderr,"  write data: %s [ options ] -w < file\n",name);
    fprintf(stderr,"\n");
    fprintf(stderr,"File format is a hex dump.  Without \"-w\" switch\n");
    fprintf(stderr,"only parsing (and printing) of the data is done,\n");
    fprintf(stderr,"writing is skipped.\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"options:\n");
    fprintf(stderr,"  -d device  device to use         [%s]\n",dev);
    fprintf(stderr,"  -a addr    set slave address     [0x%02x]\n",adr);
    fprintf(stderr,"  -x         print some hauppauge data decoded\n");
    fprintf(stderr,"\n");
    exit(1);
}

/* ---------------------------------------------------------------------- */
/* Hauppauge boards */

#define VENDOR_AVER_MEDIA 0x1461
#define VENDOR_HAUPPAUGE  0x0070
#define VENDOR_FLYVIDEO   0x1851
#define VENDOR_STB        0x10B4

#define STR(array,i) (i < sizeof(array)/sizeof(char*) ? array[i] : "unknown")

static char *chiptype[] = {
	"Reserved", "BT815", "BT817", "BT819", "BT815A", "BT817A", "BT819A",
	"BT827", "BT829", "BT848", "BT848A", "BT849A", "BT829A", "BT827A",
	"BT878", "BT879", "BT880"
};

static char *vidformats[] = {
	"NTSC(M)", "PAL(BGHIDK)", "SECAM(L/L)", "PAL(M)",
	"PAL(N)", "NTSC(443)", "PAL(NCOMBO)", "undef"};

static char *tunformats[] = {
	"unknown1", "unknown2", "PAL(B/G)", "NTSC(M)",
	"PAL(I)", "SECAM(L/L)", "PAL(D/K)", "ATSC Digital"};

static char *sndtype[] = {
	"None", "TEA6300", "TEA6320", "TDA9850", "MSP3400C", "MSP3410D",
	"MSP3415", "MSP3430", "MSP3438", "CS5331", "MSP3435", "MSP3440",
	"MSP3445", "MSP3411", "MSP3416", "MSP3425" };

static char *sndout[] = {
	"None", "Internal", "BackPanel", "Internal and BackPanel"};

static char *teletext[] = { "None", "SAA5246", "SAA5284", "Yes (Software)"};

static char *h_tuner [] = {
    "",
    "External",
    "Unspecified",
    "Phillips FI1216",
    "Phillips FI1216MF",
    "Phillips FI1236",
    "Phillips FI1246",
    "Phillips FI1256",
    "Phillips FI1216 MK2",
    "Phillips FI1216MF MK2",
    "Phillips FI1236 MK2",
    "Phillips FI1246 MK2",
    "Phillips FI1256 MK2",
    "Temic 4032FY5",
    "Temic 4002FH5",
    "Temic 4062FY5",
    "Phillips FR1216 MK2",
    "Phillips FR1216MF MK2",
    "Phillips FR1236 MK2",
    "Phillips FR1246 MK2",
    "Phillips FR1256 MK2",
    "Phillips FM1216",
    "Phillips FM1216MF",
    "Phillips FM1236",
    "Philips FM1246",
    "Philips FM1256",
    "Temic 4036FY5",
    "Samsung TCPN9082D",
    "Samsung TCPM9092P",
    "Temic 4006FH5",
    "Samsung TCPN9085D",
    "Samsung TCPB9085P",
    "Samsung TCPL9091P",
    "Temic 4039FR5",
    "Philips FQ1216 ME",
    "Temic 4066FY5",
    "Philips TD1536",
    "Philips TD1536D",
    "Philips FMR1236",
    "Philips FI1256MP",
    "Samsung TCPQ9091P",
    "Temic 4006FN5",
    "Temic 4009FR5",
    "Temic 4046FM5",
    "Temic 4009FN5",
    "Philips TD1536D_FH_44",
    "LG TP18NSR01F",
    "LG TP18PSB01D",
    "LG TP18PSB11D",
    "LG TAPC_l001D",
    "LG TAPC_l701D"
};

void dump_hauppauge(unsigned char *tvee)
{
    unsigned int model;
    unsigned int revision;
    unsigned int serial;
    unsigned int vendor_id;
    unsigned int id;
    int i, b1, b2, b3, s1, s2, s3, s4;
    int videotext, chip, tunerformats , videoformats, tuner,
	ext_inputs, svideo, radio, rds, infrared, sound, soundout,
	cksum=0;

    // BT878 reads its PCI subsystem ID at power on from eeprom
    id = (tvee[252] << 8) | tvee[253];
    vendor_id = (tvee[254] << 8) | tvee[255];
    
    printf("\nPCI Vendor ID: 0x%04x, Subsystem ID: 0x%04x\n\n", vendor_id, id);

    printf("Decoding Hauppauge EEPROM:\n");

    if (1 /* vendor_id == VENDOR_HAUPPAUGE */) {
	b1 = 2 + tvee[1];
	if ((b1+2) >255 ) {
		printf("Invalid eeprom (b1)\n");
		goto dumpep;
	}
	b2 = 2 + tvee[b1 + 2];
	if ((b1 + 2 + b2 + 1) >255) {
		printf("Invalid eeprom (b2)\n");
		goto dumpep;
	}
	b3 = 2 + tvee[b1 + 2 + b2 + 1];		/* Blocklngen - 1*/

	s1 = 0;	/* Blockstarts*/
	s2 = tvee[s1 + 1] + 3;
	s3 = tvee[s2 + 1] + 3 + s2;
	s4 = (tvee[s3] & 0x0f) +1 + s3; /* 0x74, 0x77 for TDA */
	if(s4>255) {
		printf("Invalid eeprom (s4)\n");
		goto dumpep;
	}

	// BLOCK 1:
	if(tvee[0] != 0x84 ) {
		printf("Invalid eeprom(B1).\n");
		goto dumpep;
	}
	//s1[1] is length (low byte)
	//s1[2] is length (hi byte), 0x00
	//s1[3] must be 0x00
	//s1[4] must be 0x05
	videotext= tvee[5] & 0x08; /* probably indicates 35/28 Mhz by 0x40/0x50 */
	if(videotext) {
		if(model>60000) videotext= 2;
		else videotext=1;
	}
	else videotext=3;
	chip	    = tvee[6];
	videoformats= tvee[7];
	tunerformats= tvee[8];
	tuner	    = tvee[9];
	ext_inputs  = tvee[10] &0x0f;
	svideo      =     (tvee[10] & 0x03) ? 1 : 0;
	model    = (tvee[12] << 8 | tvee[11]);
	revision = (tvee[15] << 16 | tvee[14] << 8 | tvee[13]);
	// tvee[17] is 0x00 on newer cards
	// followed by variable number of 0x00`s. The last byte
	radio    = tvee[b1] & 0x01;
	rds      = tvee[b1] & 0x02;
	infrared = tvee[b1] & 0x04;

	// BLOCK 2:
	if(tvee[s2] != 0x84) {
		printf("Invalid eeprom(B2).\n");
		goto dumpep;
	}
	//s2[1] is length (low byte)
	//s2[2] is length (hi byte), 0x00
	//s2[3..8] is constant block 01 01 20 77 00 04
	serial   = (tvee[b1 + 12] << 16 | tvee[b1 + 11] << 8 | tvee[b1 + 10]);
	//s2[12] is by 0x00

	// BLOCK 3:
	if(((tvee[s3])& 0xf0) != 0x70) {
		 printf("Invalid eeprom(B3).\n");
		goto dumpep;
	}
	// s3[1] can interact with tuner/video formats (normal 0x02)
	// s3[2] must be 0x01
	sound	    = tvee[s3 + 3] &0x0f;
	soundout    = tvee[s3 + 4];

	// BLOCK 4:
	if(tvee[s4] != 0x79) {
		printf("Invalid eeprom(B4).\n");
		goto dumpep;
	}
	// s4[1] is checksum byte
	for(i=0; i<s4+2; i++)
		cksum=cksum + tvee[i];
	cksum = cksum &0xff;

	printf("Model        : %d\n", model);
	printf("Revision     : %c%c%c%c\n",
	       ((revision >> 18) & 0x3f) + 32,
	       ((revision >> 12) & 0x3f) + 32,
	       ((revision >> 6) & 0x3f) + 32,
	       ((revision >> 0) & 0x3f) + 32);
	printf("Serial       : %d\n", serial);
	printf("Tuner        : %s\n", h_tuner[tuner]);
	printf("Tuner Formats:");
	for(i=0;i<8;i++)
		if(tunerformats & (1<<i)) printf(" %s",tunformats[i]);
	printf("\n");
	printf("Audio        : %s\n", sndtype[sound]);
	printf("Audio Outputs: %s\n", sndout[soundout]);
	printf("Decoder      : %s\n", chiptype[tvee[6]]);
	printf("Video Formats:");
	for(i=0;i<8;i++)
		if(videoformats & (1<<i)) printf(" %s",vidformats[i]);
	printf("\n");

	printf("Videotext    : %s\n", teletext[videotext]);
	printf("Radio        : %s%s\n", radio ?"FM":"no" , rds?" (RDBS)":"");
	printf("Infrared     : %s\n",infrared?"yes":"no");

	printf("Checksum     : %x",tvee[s4+1]);
	if(cksum == 0)	
		printf(" [OK]\n");
	else 
		printf(" [INVALID, should be %x]\n", 
		(0x100 + tvee[s4+1] - cksum)& 0xff);


	dumpep:
	printf("\nEEPROM:\n");
	for (i = 0; i <= 40; i++) {
	    if (i == (b1 + 1) || i == (b1 + b2 + 2) 
		|| i == (b1 + b2 + b3 + 3))
		printf("\n");
	    printf("%02x ", tvee[i]);
	}
	printf("\n");
    }
}
    
/* ---------------------------------------------------------------------- */

void dump_buf(unsigned char *buf)
{
    int i,j;
    
    for (i = 0; i < 256; i += 16) {
	printf("%04x  ",i);
	for (j = i; j < i+16; j++) {
	    if (!(j%4))
		printf(" ");
	    printf("%02x ",buf[j]);
	}
	printf("  ");
	for (j = i; j < i+16; j++)
	    printf("%c",isalnum(buf[j]) ? buf[j] : '.');
	printf("\n");
    }
}

int parse_buf(unsigned char *buf)
{
    int  i,j,n,pos,count;
    char line[100];
    
    for (i = 0; i < 256; i += 16) {
	if (NULL == fgets(line,99,stdin)) {
	    fprintf(stderr,"unexpected EOF\n");
	    return -1;
	}
	if ('#' == line[0] || '\n' == line[0]) {
            i -= 16;
	    continue;
	}
	if (1 != sscanf(line,"%x%n",&n,&pos)) {
	    fprintf(stderr,"addr parse error (%d)\n",i>>4);
	    return -1;
	}
	if (n != i) {
	    fprintf(stderr,"addr mismatch\n");
	    return -1;
	}
	for (j = i; j < i+16; j++) {
	    if (1 != sscanf(line+pos,"%x%n",&n,&count)) {
		fprintf(stderr,"value parse error\n");
		return -1;
	    }
	    buf[j] = n;
	    pos += count;
	}
    }
    return 0;
}

/* ---------------------------------------------------------------------- */

void
read_buf(int fd, unsigned char *buf)
{
    int i,n=8; 
    unsigned char addr;

#if 0
    addr = 0;
    if (-1 == (write(fd,&addr,1))) {
	fprintf(stderr,"write addr %s: %s\n",dev,strerror(errno));
	exit(1);
    }
#endif
    for (i = 0; i < 256; i += n) {
#if 1
	addr = i;
	if (-1 == (write(fd,&addr,1))) {
	    fprintf(stderr,"write addr %s: %s\n",dev,strerror(errno));
	    exit(1);
	}
#endif
	if (-1 == (read(fd,buf+i,n))) {
	    fprintf(stderr,"read data %s: %s\n",dev,strerror(errno));
	    exit(1);
	}
    }
}

void
write_buf(int fd, unsigned char *buf)
{
    int i,j,n = 8; 
    unsigned char tmp[17];
    
    for (i = 0; i < 256; i += n) {
	tmp[0] = i;
	for (j = 0; j < n; j++)
	    tmp[j+1] = buf[i+j];
	if (-1 == (write(fd,tmp,n+1))) {
	    fprintf(stderr," write data %s: %s\n",dev,strerror(errno));
	    exit(1);
	}
	fprintf(stderr,"*");
	usleep(100000); /* 0.1 sec */
    }
}

/* ---------------------------------------------------------------------- */

int main(int argc, char *argv[])
{
    int            f,c;
    unsigned char  buf[256];
    
    /* parse options */
    opterr=1;
    while ( (c=getopt(argc,argv,"hwxa:d:")) != -1) {
	switch (c){
	case 'w':
	    wr=1;
	    break;
	case 'x':
	    x=1;
	    break;
	case 'd':
	    dev = optarg;
	    break;
	case  'a':
	    adr = strtol(optarg,NULL,0);
	    break;
	case 'h':
	default:
	    usage(argv[0]);
	}
    }

    if (-1 == (f = open(dev,O_RDWR))) {
	fprintf(stderr,"open %s: %s\n",dev,strerror(errno));
	exit(1);
    }
    ioctl(f,I2C_SLAVE,adr>>1);
    memset(buf,0,256);

    if (isatty(fileno(stdin))) {
	/* read */
	read_buf(f,buf);
	dump_buf(buf);
	if (x)
	    dump_hauppauge(buf);
    } else {
	/* write */
	if (-1 == parse_buf(buf))
	    exit(1);
	dump_buf(buf);
	if (wr) {
	    fprintf(stderr,"writing to eeprom now... ");
	    write_buf(f,buf);
	    fprintf(stderr," ok\n");
	}
    }
    
    close(f);
    exit(0);
}
